package org.openmuc.jdlms.internal.lnassociation;

import java.io.IOException;
import java.util.List;

import org.openmuc.jdlms.IllegalAttributeAccessException;
import org.openmuc.jdlms.ObisCode;
import org.openmuc.jdlms.SelectiveAccessDescription;
import org.openmuc.jdlms.datatypes.DataObject;
import org.openmuc.jdlms.internal.APdu;
import org.openmuc.jdlms.internal.DataConverter;
import org.openmuc.jdlms.internal.asn1.axdr.types.AxdrEnum;
import org.openmuc.jdlms.internal.asn1.axdr.types.AxdrOptional;
import org.openmuc.jdlms.internal.asn1.cosem.COSEMpdu;
import org.openmuc.jdlms.internal.asn1.cosem.Cosem_Attribute_Descriptor;
import org.openmuc.jdlms.internal.asn1.cosem.Cosem_Attribute_Descriptor_With_Selection;
import org.openmuc.jdlms.internal.asn1.cosem.GET_Request;
import org.openmuc.jdlms.internal.asn1.cosem.GET_Response;
import org.openmuc.jdlms.internal.asn1.cosem.Get_Data_Result;
import org.openmuc.jdlms.internal.asn1.cosem.Get_Request_Normal;
import org.openmuc.jdlms.internal.asn1.cosem.Get_Request_With_List;
import org.openmuc.jdlms.internal.asn1.cosem.Get_Response_Normal;
import org.openmuc.jdlms.internal.asn1.cosem.Get_Response_With_List;
import org.openmuc.jdlms.internal.asn1.cosem.Get_Response_With_List.SubSeqOf_result;
import org.openmuc.jdlms.internal.asn1.cosem.Invoke_Id_And_Priority;
import org.openmuc.jdlms.internal.asn1.cosem.Selective_Access_Descriptor;

public class GetRequestProcessor implements RequestProcessor {

    private final AssociationMessenger associationMessenger;
    private final RequestProcessorData requestProcessorData;

    public GetRequestProcessor(AssociationMessenger associationMessenger, RequestProcessorData requestProcessorData) {
        this.associationMessenger = associationMessenger;
        this.requestProcessorData = requestProcessorData;
    }

    @Override
    public void processRequest(COSEMpdu request) throws IOException {
        GET_Request getRequest = request.get_request;
        GET_Response getResponse;

        switch (getRequest.getChoiceIndex()) {
        case GET_REQUEST_NORMAL:
            getResponse = processGetRequestNormal(getRequest.get_request_normal);
            break;

        case GET_REQUEST_WITH_LIST:
            getResponse = processGetRequestWithList(getRequest.get_request_with_list);
            break;

        case GET_REQUEST_NEXT:
            throw new IOException("not yet implemented");

        default:
        case _ERR_NONE_SELECTED:
            // TODO
            throw new IOException("should no happen");
        }

        COSEMpdu coseMpdu = new COSEMpdu();
        coseMpdu.setget_response(getResponse);

        APdu aPdu = new APdu(null, coseMpdu);

        byte[] data = this.associationMessenger.encodeApdu(aPdu);

        long clientMaxReceivePduSize = this.requestProcessorData.connectionData.clientMaxReceivePduSize;
        if (clientMaxReceivePduSize != 0 && data.length > clientMaxReceivePduSize) {
            splitGetResponse(data);
        }

        this.associationMessenger.send(data);
    }

    private GET_Response processGetRequestWithList(Get_Request_With_List requestWithList) {
        Invoke_Id_And_Priority invokeIdAndPriority = requestWithList.invoke_id_and_priority;

        List<Cosem_Attribute_Descriptor_With_Selection> list = requestWithList.attribute_descriptor_list.list();

        SubSeqOf_result result = new SubSeqOf_result();
        for (Cosem_Attribute_Descriptor_With_Selection attributeDescriptor : list) {

            Get_Data_Result element = tryGet(attributeDescriptor.cosem_attribute_descriptor,
                    attributeDescriptor.access_selection);
            result.add(element);
        }

        GET_Response getResponse = new GET_Response();
        Get_Response_With_List responseWithList = new Get_Response_With_List(invokeIdAndPriority, result);
        getResponse.setget_response_with_list(responseWithList);

        return getResponse;
    }

    private GET_Response processGetRequestNormal(Get_Request_Normal normalRequest) {
        Invoke_Id_And_Priority invokeIdAndPriority = normalRequest.invoke_id_and_priority;

        Get_Data_Result result = tryGet(normalRequest.cosem_attribute_descriptor, normalRequest.access_selection);
        GET_Response getResponse = new GET_Response();
        getResponse.setget_response_normal(new Get_Response_Normal(invokeIdAndPriority, result));

        return getResponse;
    }

    private Get_Data_Result tryGet(Cosem_Attribute_Descriptor cosemAttributeDescriptor,
            AxdrOptional<Selective_Access_Descriptor> accessSelection) {
        Get_Data_Result result = new Get_Data_Result();
        try {
            ObisCode instanceId = new ObisCode(cosemAttributeDescriptor.instance_id.getValue());

            SelectiveAccessDescription selectiveAccessDescription = null;
            if (accessSelection.isUsed()) {
                Selective_Access_Descriptor accessDescriptor = accessSelection.getValue();

                int accessSelector = (int) accessDescriptor.access_selector.getValue();
                DataObject accessParameter = DataConverter.convertDataToDataObject(accessDescriptor.access_parameters);
                selectiveAccessDescription = new SelectiveAccessDescription(accessSelector, accessParameter);
            }

            DataObject object = this.requestProcessorData.directory.get(requestProcessorData.logicalDeviceId,
                    instanceId, cosemAttributeDescriptor.class_id.getValue(),
                    cosemAttributeDescriptor.attribute_id.getValue(), selectiveAccessDescription,
                    this.requestProcessorData.connectionId);

            result.setdata(DataConverter.convertDataObjectToData(object));
        } catch (IllegalAttributeAccessException e) {
            result.setdata_access_result(new AxdrEnum(e.getAccessResultCode().getCode()));
        }
        return result;
    }

    private void splitGetResponse(byte[] data) throws IOException {
        // TODO: implement this..!
        throw new IOException("Not yet implemented");
    }

}
