package org.openmuc.jdlms;

import java.io.IOException;
import java.io.Serializable;
import java.text.MessageFormat;
import java.util.Map;

/**
 * Object info for the short name addressing.
 */
public class SnObjectInfo implements Serializable {
    private final int baseName;
    private final int classId;
    private final int version;

    /**
     * Create a new short name object info.
     * 
     * @param baseName
     *            the base name (SN Address).
     * @param classId
     *            the class id of the COSEM object.
     * @param version
     *            the version of the COSEM object.
     */
    public SnObjectInfo(int baseName, int classId, int version) {
        this.baseName = baseName;
        this.classId = classId;
        this.version = version;
    }

    /**
     * Get the base name (SN address) of a COSEM objec.
     * 
     * @return the base name.
     */
    public int getBaseName() {
        return baseName;
    }

    /**
     * Get the class COSEM ID.
     * 
     * @return the class ID.
     */
    public int getClassId() {
        return classId;
    }

    /**
     * Get the COSEM class version.
     * 
     * @return the version.
     */
    public int getVersion() {
        return version;
    }

    @Override
    public String toString() {
        return MessageFormat.format("Base Name: {0}; Class ID: {1}; Version: {2}", this.baseName, this.classId,
                this.version);
    }

    /**
     * Retrieve the LN to SN mapping from the current connection.
     * 
     * @param connection
     *            a DlmsConnection using short naming.
     * @return the LN -&gt; SN mapping.
     * @throws IOException
     *             if the connection is no a short name connection or an error occurs.
     */
    public static Map<ObisCode, SnObjectInfo> retrieveLnSnMappingFrom(DlmsConnection connection) throws IOException {
        if (!(connection instanceof DlmsSnConnection)) {
            throw new IOException("This operation in only available if you're using an SN connection.");
        }

        DlmsSnConnection snConnection = (DlmsSnConnection) connection;

        return snConnection.getLatestObjectInfoMapping();
    }
}
