/*
 * Copyright 2012-16 Fraunhofer ISE
 *
 * This file is part of jDLMS.
 * For more information visit http://www.openmuc.org
 *
 * jDLMS is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * jDLMS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with jDLMS.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
package org.openmuc.jdlms.app.client;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.TimeoutException;

import org.openmuc.jdlms.AccessResultCode;
import org.openmuc.jdlms.AttributeAddress;
import org.openmuc.jdlms.DlmsConnection;
import org.openmuc.jdlms.GetResult;
import org.openmuc.jdlms.ObisCode;
import org.openmuc.jdlms.ResponseTimeoutException;
import org.openmuc.jdlms.datatypes.DataObject;
import org.openmuc.jdlms.interfaceclass.InterfaceClass;
import org.openmuc.jdlms.interfaceclass.attribute.AssociationSnAttribute;

class SnActionProcessor extends GenActionProcessor {
    private static final String LOGICAL_NAME_FORMAT = "<Interface_Class_ID>/<OBIS_Code>/<Object_Attribute_ID>";

    private final DlmsConnection connection;

    private boolean closed;

    public SnActionProcessor(DlmsConnection connection) {
        this.connection = connection;
        this.closed = false;
    }

    @Override
    public void close() {
        try {
            if (!closed) {
                connection.close();
                closed = true;
            }
        } catch (IOException e) {
            System.out.println("Error occured while closing.");
        }
    }

    @Override
    protected String nameFormat() {
        return LOGICAL_NAME_FORMAT;
    }

    @Override
    protected GetResult callGet(String requestParameter)
            throws IOException, TimeoutException, IllegalArgumentException {
        AttributeAddress attributeAddress;

        try {
            attributeAddress = buidlAttributeAddress(requestParameter);
        } catch (IllegalArgumentException e) {
            // TODO: handle this better
            // System.err.println(e.getMessage());
            // return;
            throw e;
        }
        return connection.get(attributeAddress);
    }

    private AttributeAddress buidlAttributeAddress(String requestParameter)
            throws IllegalArgumentException, NumberFormatException {
        String[] arguments = requestParameter.split("/");

        if (arguments.length != 3) {
            throw new IllegalArgumentException(String.format("Wrong number of arguments. %s", LOGICAL_NAME_FORMAT));
        }

        int classId = Integer.parseInt(arguments[0]);

        ObisCode obisCode = new ObisCode(arguments[1]);

        int attributeId = Integer.parseInt(arguments[2]);

        return new AttributeAddress(classId, obisCode, attributeId);
    }

    //@formatter:off
    /*
     * object_list
     * 
     * object_list_type ::= array   objlist_element
     * objlist_element  ::= structure
     * {
     *                      base_name:              long,
     *                      class_id:               long-unsigned,
     *                      version:                unsigned,
     *                      logical_name:   octet-string
     * }
     * 
     * 
     * 
     * access_rights_list
     * 
     * access_rights_element ::= structure
     * {
     *              base_name:                      long,
     *              attribute_access:       attribute_access_descriptor,
     *              method_access:          method_access_descriptor
     * }
     * 
     * attribute_access_descriptor ::= array        attribute_access_item
     * 
     * attribute_access_item ::= structure
     * {
     *              attribute_id:   integer,
     *              access_mode:    enum,
     *                                              {
     *                                                      (0) no_access,
     *                                                      (1) read_only,
     *                                                      (2) write_only,
     *                                                      (3) read_and_write,
     *                                                      (4) authenticated_read_only,
     *                                                      (5) authenticated_write_only,
     *                                                      (6) authenticated_read_and_write
     *                                              }
     *              access_selectors: CHOICE
     *                                              {
     *                                                      null-data,
     *                                                      array of integer
     *                                              }
     * }
     * 
     * method_access_descriptor ::= array   method_access_item
     * 
     * method_access_item ::= structure
     * {
     *              method_id: integer,
     *              access_mode: enum
     *                                      {
     *                                              (0) no_access,
     *                                              (1) access,
     *                                              (2) authenticated_access
     *                                      }
     * }
     */
     //@formatter:on

    @Override
    public void processScanObjects() throws IOException {
        AttributeAddress snObjects = new AttributeAddress(AssociationSnAttribute.OBJECT_LIST,
                new ObisCode("0.0.40.0.0.255"));
        AttributeAddress snAccessRights = new AttributeAddress(AssociationSnAttribute.ACCESS_RIGHTS_LIST,
                new ObisCode("0.0.40.0.0.255"));

        GetResult objectsResult;
        GetResult accessRightsResult;
        try {
            List<GetResult> list = connection.get(Arrays.asList(snObjects, snAccessRights));
            objectsResult = list.get(0);
            accessRightsResult = list.get(1);
        } catch (ResponseTimeoutException e) {
            System.err.println("Failed to scan: Timeout");
            return;
        }

        if (objectsResult.getResultCode() != AccessResultCode.SUCCESS) {
            System.err.printf("Failed to scan with following result code %s.\n", objectsResult.getResultCode().name());
            return;
        }
        System.out.println(accessRightsResult.getResultCode());

        List<DataObject> objectList = objectsResult.getResultData().getValue();

        final String formatStr = "%-30s %-10s %s\n";
        System.out.printf(formatStr, "Class_ID", "Short Name", "Obis Code");
        for (DataObject objectListElement : objectList) {
            List<DataObject> struct = objectListElement.getValue();

            Number baseName = struct.get(0).getValue();
            Number classId = struct.get(1).getValue();
            classId = classId.intValue() & 0xFFFFFFFFL;
            Number version = struct.get(2).getValue();

            byte[] logNameBytes = struct.get(3).getValue();

            ObisCode obisCode = new ObisCode(logNameBytes);

            InterfaceClass interfaceClass = InterfaceClass.interfaceClassFor(classId.intValue(), version.intValue());

            String interfaceClassStr;
            if (interfaceClass != InterfaceClass.UNKNOWN) {
                interfaceClassStr = String.format("%s(%d)", interfaceClass.name(), classId);
            }
            else {
                interfaceClassStr = classId.toString();
            }

            System.out.printf(formatStr, interfaceClassStr, String.format("%04X", baseName.intValue() & 0xFFFF),
                    obisCode.toDecimal());
        }
    }

    @Override
    protected AccessResultCode callSet(String requestParameter, DataObject dataToWrite) throws IOException {
        throw new UnsupportedOperationException("Set operation is not yet implemented for SN console APP..");
    }

}
